/*-
 * Copyright (c) 2007 Benedikt Stockebrand
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "tcp_open.h"

/* Open a TCP client socket to a given server/IP
   address at a given service/port number.  If
   custom_hints is non-NULL, use it instead of the
   default hints for the cetaddrinfo(3) call.  If
   gai_error is non-NULL, store getaddrinfo(3)
   error return codes there. */

int tcp_open(const char            * server, 
	     const char            * service,
	     const struct addrinfo * custom_hints,
	     int                   * gai_error)
{
  struct addrinfo   deflt_hints;
  struct addrinfo * results;
  int               status;
  struct addrinfo * i;
  int               sockd;
  
  /* Initialize it right away so we don't have to
     bother about it again later on */
  if (gai_error) *gai_error = 0;

  /* Do the getaddrinfo(3) invocation */
  if (custom_hints == NULL) {
    memset((void *) &deflt_hints, 0, 
	   sizeof(deflt_hints));
    deflt_hints.ai_family   = PF_UNSPEC;
    deflt_hints.ai_socktype = SOCK_STREAM;
    status = getaddrinfo(server, service, 
			 &deflt_hints, &results);
  }
  else
    status = getaddrinfo(server, service, 
			 custom_hints, &results);
  if (status) {
    if (gai_error) *gai_error = status;
    return(-1);
  }
  
  /* Walk the result list until a connection gets
     established */
  for (i=results; i; i=i->ai_next)
    {
      /* Create a socket---this shouldn't fail
	 except if there is a resource
	 scarcity. If the result list is long and
	 a large number of unsuccessful connection
	 attempts occur, then this is suboptimal;
	 adding a "cache" for pre-allocated
	 sockets will however add significant
	 complexity to the code and make it
	 vulnerable to bad API implementations
	 that can't re-use a socket if a previous
	 connect(2) has failed. */
      sockd = socket(i->ai_family, 
		     i->ai_socktype, 
		     i->ai_protocol);
      if (sockd < 0) return(-1);

      /* Try to connect to the server */
      status = connect(sockd, i->ai_addr, 
		       i->ai_addrlen);
      if (status == 0) {
	/* Success, clean up getaddrinfo data and
	   return socket */
	freeaddrinfo(results);
	return(sockd);
      }
      /* Failure, release allocated socket and try
	 next */
      status = close(sockd);
      if (status) return(-1);
    }

  /* None of the getaddrinfo(3) results worked;
     give up */
  return(-1);
}

/* Local Variables:
eval: (c-set-style "gnu")
eval: (set-fill-column 50)
*/
