/*-
 * Copyright (c) 2007 Benedikt Stockebrand
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <signal.h>
#include <sys/wait.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#include "tcp_server.h"

/* Whenever a child process terminates, the main
   process will receive a SIGCHLD.  The child
   process will remain a zombie until waitpid(2)
   is called, which is what this signal handler
   does.
*/
void handle_sigchld(int signo);


void tcp_server(const char            * service, 
		const struct addrinfo * custom_hints,
		connection_handler    * handler)
{
  struct addrinfo   hints;
  struct addrinfo * results;
  int               status;
  int               srv_sockd;
  int               sock_opt;

  /* Use getaddrinfo(3) to get the socket
     configuration */
  if (custom_hints) 
    memcpy(&hints, custom_hints, sizeof(hints));
  else
    memset(&hints, 0, sizeof(hints));
  hints.ai_flags |= AI_PASSIVE;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_protocol = IPPROTO_TCP;
  if (hints.ai_family == PF_UNSPEC)
    hints.ai_family = PF_INET6;
  status = getaddrinfo(NULL, service, 
		       &hints, &results);
  if (status) {
    syslog(LOG_ERR, "getaddrinfo(3): %s",
	   gai_strerror(status));
    exit(EXIT_FAILURE);
  }
  if (results == NULL) {
    syslog(LOG_ERR, "getaddrinfo(3): No results");
    exit(EXIT_FAILURE);
  }

  /* Allocate the server socket with socket(2) */
  srv_sockd = socket (results->ai_family, 
		      results->ai_socktype,
		      results->ai_protocol);
  if (srv_sockd < 0) {
    syslog(LOG_ERR, "socket(2): %s",
	   strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* Set the SO_REUSEADDR socket option */
  sock_opt = 1;
  status = setsockopt(srv_sockd, 
		      SOL_SOCKET, SO_REUSEADDR, 
		      (char *) &sock_opt, 
		      sizeof(sock_opt));
  if (status) {
    syslog(LOG_ERR, "setsockopt(2): %s",
	   strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* Set the IPV6_V6ONLY socket option */
  if (custom_hints && custom_hints->ai_family 
      == PF_INET6)
    sock_opt = 1;
  else
    sock_opt = 0;
  status = setsockopt(srv_sockd, 
		      IPPROTO_IPV6, IPV6_V6ONLY, 
		      (char *) &sock_opt, 
		      sizeof(sock_opt));
  if (status) {
    syslog(LOG_ERR, "setsockopt(2): %s",
	   strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  /* Bind the local side of the server socket with
     bind(2) */
  status = bind(srv_sockd, results->ai_addr,
		results->ai_addrlen);
  if (status) {
    syslog(LOG_ERR, "bind(2): %s",
	   strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  /* Call listen(2) to tell the OS that this is a
     server socket */
  status = listen(srv_sockd, 255);
  if (status) {
    syslog(LOG_ERR, "listen(2): %s",
	   strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* Initialize the SIGCHLD handler */
  signal(SIGCHLD, handle_sigchld);
  
  /* Daemonize */
  daemon(0, 0);
  
  /* Do the accept(2) loop.  If anything goes
     wrong within the loop, log the error, close
     the connection socket and kill the
     client-handling process but keep the main
     process running at all costs. */
  while (1) {
    int                     clt_sockd;
    struct sockaddr_storage clt_sa;
    socklen_t               clt_sa_len;
    pid_t                   child_pid;

    /* Accept(2) an incoming connection */
    clt_sa_len=sizeof(clt_sa);
    clt_sockd 
      = accept(srv_sockd, 
	       (struct sockaddr *) &clt_sa,
	       &clt_sa_len);
    if (clt_sockd < 0) {
      if (errno != EINTR)
	syslog(LOG_ERR, "accept(2): %s",
	       strerror(errno));
      continue;
    }

    /* Fork(2) the connection handling process */
    child_pid = fork();
    if (child_pid < 0) {
      /* The fork(2) call failed, now try to clean
	 up */
      syslog(LOG_ERR, "fork(2): %s",
	     strerror(errno));
      status = close(clt_sockd);
      if (status)
	syslog(LOG_ERR, "close(2): %s",
	       strerror(errno));
    }
    else if (child_pid > 0) {
      /* Main process: close connection socket */
      status = close(clt_sockd);
      if (status)
	syslog(LOG_ERR, "close(2): %s",
	       strerror(errno));
    }
    else {
      /* Child process: close server socket */
      status = close(srv_sockd);
      if (status) {
	syslog(LOG_ERR, 
	       "(PID %d) close(2): %s (killed)",
	       getpid(), strerror(errno));
	exit(EXIT_FAILURE);
      }

      /* Child process: call connection handler */
      handler(clt_sockd,
	      (struct sockaddr *) &clt_sa,
	      clt_sa_len);
      exit(EXIT_SUCCESS);
    }
  }
}

void handle_sigchld(int signo)
{
  int status;
  while (waitpid(-1, &status, WNOHANG) > 0)
    ;
  return;
}


/* Local Variables:
eval: (c-set-style "gnu")
eval: (set-fill-column 76)
*/
